"""Computation of CD0 for a lifting surface."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

import numpy as np

from .friction_drag import get_flat_plate_friction_drag_coefficient


@dataclass
class LiftingSurfaceGeometry:
    """Minimum geometry data for computation of CD0 of lifting surfaces."""

    thickness_ratio: float  #: average thickness ratio
    MAC_length: float  #: length of Mean Aerodynamic Chord
    sweep_angle_25: float  #: sweep angle at 25% chord, in degrees
    cambered: bool  #: True if airfoil is cambered
    wet_area: float  #: wet surface area of the lifting surface
    interaction_coeff: float  #: ratio of additional drag due to interaction effects


def compute_cd0_lifting_surface(
    geometry: LiftingSurfaceGeometry,
    mach: float,
    reynolds: float,
    wing_area: float,
    lift_coefficient: float = 0.0,
):
    """
    Computes CD0 for a lifting surface.

    Friction coefficient is assessed from :cite:`raymer:1999` (Eq 12.27).
    Corrections for lifting surfaces are from :cite:`supaero:2014`.

    :param geometry: definition of lifting surface geometry
    :param mach: Mach number
    :param reynolds: Reynolds number
    :param wing_area: wing area (will be used for getting CD specific to wing area
    :param lift_coefficient: needed if wing is cambered
    :return: CD0 value
    """
    # Drag coefficient for flat plate (
    cf = get_flat_plate_friction_drag_coefficient(geometry.MAC_length, mach, reynolds)

    # Contribution of relative thickness
    thickness_contribution = 4.688 * geometry.thickness_ratio**2 + 3.146 * geometry.thickness_ratio

    # Contribution of camber
    sweep_25 = np.radians(geometry.sweep_angle_25)
    if geometry.cambered:
        camber_contribution = (
            2.859 * (lift_coefficient / np.cos(sweep_25) ** 2) ** 3
            - 1.849 * (lift_coefficient / np.cos(sweep_25) ** 2) ** 2
            + 0.382 * (lift_coefficient / np.cos(sweep_25) ** 2)
            + 0.06
        )
    else:
        camber_contribution = 0.0

    # Correction from sweep angle
    sweep_correction = 1 - 0.000178 * geometry.sweep_angle_25**2 - 0.0065 * geometry.sweep_angle_25

    cd0_ht = (
        (
            (thickness_contribution + camber_contribution) * sweep_correction
            + geometry.interaction_coeff
            + 1
        )
        * cf
        * geometry.wet_area
        / wing_area
    )
    return cd0_ht



"""
Estimation of horizontal tail mean aerodynamic chords
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import math

import numpy as np
import openmdao.api as om


# TODO: it would be good to have a function to compute MAC for HT, VT and WING
class ComputeHTMAC(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Horizontal tail mean aerodynamic chord estimation"""

    def setup(self):
        self.add_input("data:geometry:horizontal_tail:center:chord", val=np.nan, units="m")
        self.add_input("data:geometry:horizontal_tail:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:horizontal_tail:sweep_25", val=np.nan, units="rad")
        self.add_input("data:geometry:horizontal_tail:span", val=np.nan, units="m")

        self.add_output("data:geometry:horizontal_tail:MAC:length", units="m")
        self.add_output("data:geometry:horizontal_tail:MAC:at25percent:x:local", units="m")
        self.add_output("data:geometry:horizontal_tail:MAC:y", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:horizontal_tail:MAC:length",
            [
                "data:geometry:horizontal_tail:center:chord",
                "data:geometry:horizontal_tail:tip:chord",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:horizontal_tail:MAC:at25percent:x:local",
            [
                "data:geometry:horizontal_tail:center:chord",
                "data:geometry:horizontal_tail:tip:chord",
                "data:geometry:horizontal_tail:sweep_25",
                "data:geometry:horizontal_tail:span",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:horizontal_tail:MAC:y",
            [
                "data:geometry:horizontal_tail:center:chord",
                "data:geometry:horizontal_tail:tip:chord",
                "data:geometry:horizontal_tail:span",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        root_chord = inputs["data:geometry:horizontal_tail:center:chord"]
        tip_chord = inputs["data:geometry:horizontal_tail:tip:chord"]
        sweep_25_ht = inputs["data:geometry:horizontal_tail:sweep_25"]
        b_h = inputs["data:geometry:horizontal_tail:span"]

        mac_ht = (
            (root_chord**2 + root_chord * tip_chord + tip_chord**2)
            / (tip_chord + root_chord)
            * 2
            / 3
        )

        x0_ht = (root_chord / 4) + (b_h / 6) * math.tan(sweep_25_ht) * (
            (root_chord + 2 * tip_chord) / (root_chord + tip_chord)
        )

        y0_ht = (b_h * (0.5 * root_chord + tip_chord)) / (3 * (root_chord + tip_chord))

        outputs["data:geometry:horizontal_tail:MAC:length"] = mac_ht
        outputs["data:geometry:horizontal_tail:MAC:at25percent:x:local"] = x0_ht
        outputs["data:geometry:horizontal_tail:MAC:y"] = y0_ht



"""Computation of airframe mass."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import (
    SERVICE_EMPENNAGE_MASS,
    SERVICE_FLIGHT_CONTROLS_MASS,
    SERVICE_FUSELAGE_MASS,
    SERVICE_LANDING_GEARS_MASS,
    SERVICE_PAINT_MASS,
    SERVICE_PYLONS_MASS,
    SERVICE_WING_MASS,
)
from ..constants import SERVICE_AIRFRAME_MASS, SERVICE_GUST_LOADS


@RegisterSubmodel(SERVICE_AIRFRAME_MASS, "fastoad.submodel.weight.mass.airframe.legacy")
class AirframeWeight(om.Group):
    """
    Computes mass of airframe.
    """

    def setup(self):
        # Airframe
        self.add_subsystem(
            "loads", RegisterSubmodel.get_submodel(SERVICE_GUST_LOADS), promotes=["*"]
        )
        self.add_subsystem(
            "wing_weight", RegisterSubmodel.get_submodel(SERVICE_WING_MASS), promotes=["*"]
        )
        self.add_subsystem(
            "fuselage_weight", RegisterSubmodel.get_submodel(SERVICE_FUSELAGE_MASS), promotes=["*"]
        )
        self.add_subsystem(
            "empennage_weight",
            RegisterSubmodel.get_submodel(SERVICE_EMPENNAGE_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "flight_controls_weight",
            RegisterSubmodel.get_submodel(SERVICE_FLIGHT_CONTROLS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "landing_gear_weight",
            RegisterSubmodel.get_submodel(SERVICE_LANDING_GEARS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "pylons_weight", RegisterSubmodel.get_submodel(SERVICE_PYLONS_MASS), promotes=["*"]
        )
        self.add_subsystem(
            "paint_weight", RegisterSubmodel.get_submodel(SERVICE_PAINT_MASS), promotes=["*"]
        )

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:airframe:mass",
            [
                "data:weight:airframe:wing:mass",
                "data:weight:airframe:fuselage:mass",
                "data:weight:airframe:horizontal_tail:mass",
                "data:weight:airframe:vertical_tail:mass",
                "data:weight:airframe:flight_controls:mass",
                "data:weight:airframe:landing_gear:main:mass",
                "data:weight:airframe:landing_gear:front:mass",
                "data:weight:airframe:pylon:mass",
                "data:weight:airframe:paint:mass",
            ],
            units="kg",
            desc="Mass of airframe",
        )

        self.add_subsystem("airframe_weight_sum", weight_sum, promotes=["*"])



"""Computation of form drag for Vertical Tail Plane."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .utils.cd0_lifting_surface import (
    LiftingSurfaceGeometry,
    compute_cd0_lifting_surface,
)
from ..constants import SERVICE_CD0_VERTICAL_TAIL


@RegisterSubmodel(
    SERVICE_CD0_VERTICAL_TAIL, "fastoad.submodel.aerodynamics.CD0.vertical_tail.legacy"
)
class Cd0VerticalTail(om.ExplicitComponent):
    """
    Computation of form drag for Vertical Tail Plane.

    See :meth:`~fastoad_cs25.models.aerodynamics.components.utils.cd0_lifting_surface`
    for used method.
    """

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.add_input("data:geometry:vertical_tail:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:thickness_ratio", val=np.nan)
        self.add_input("data:geometry:vertical_tail:sweep_25", val=np.nan, units="deg")
        self.add_input("data:geometry:vertical_tail:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        if self.options["low_speed_aero"]:
            self.add_input("data:aerodynamics:wing:low_speed:reynolds", val=np.nan)
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:vertical_tail:low_speed:CD0")
        else:
            self.add_input("data:aerodynamics:wing:cruise:reynolds", val=np.nan)
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_output("data:aerodynamics:vertical_tail:cruise:CD0")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        vt_geometry = LiftingSurfaceGeometry(
            thickness_ratio=inputs["data:geometry:vertical_tail:thickness_ratio"],
            MAC_length=inputs["data:geometry:vertical_tail:MAC:length"],
            sweep_angle_25=inputs["data:geometry:vertical_tail:sweep_25"],
            wet_area=inputs["data:geometry:vertical_tail:wetted_area"],
            cambered=False,
            interaction_coeff=0.005,
        )
        wing_area = inputs["data:geometry:wing:area"]
        if self.options["low_speed_aero"]:
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]
            reynolds = inputs["data:aerodynamics:wing:low_speed:reynolds"]
        else:
            mach = inputs["data:TLAR:cruise_mach"]
            reynolds = inputs["data:aerodynamics:wing:cruise:reynolds"]

        cd0_vt = compute_cd0_lifting_surface(vt_geometry, mach, reynolds, wing_area)

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:vertical_tail:low_speed:CD0"] = cd0_vt
        else:
            outputs["data:aerodynamics:vertical_tail:cruise:CD0"] = cd0_vt



"""
Aero computation for landing phase
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterSubmodel
from stdatm import Atmosphere

from .constants import (
    SERVICE_HIGH_LIFT,
    SERVICE_LANDING_MACH_REYNOLDS,
    SERVICE_LANDING_MAX_CL,
    SERVICE_LANDING_MAX_CL_CLEAN,
    SERVICE_XFOIL,
)
from .external.xfoil.xfoil_polar import (
    OPTION_ALPHA_END,
    OPTION_ALPHA_START,
    OPTION_ITER_LIMIT,
    OPTION_XFOIL_EXE_PATH,
)


@RegisterOpenMDAOSystem("fastoad.aerodynamics.landing.legacy", domain=ModelDomain.AERODYNAMICS)
class AerodynamicsLanding(om.Group):
    """
    Computes aerodynamic characteristics at landing.

    - Computes CL and CD increments due to high-lift devices at landing.
    - Computes maximum CL of the aircraft in landing conditions.

    Maximum 2D CL without high-lift is computed using XFoil (or provided as input if option
    use_xfoil is set to False). 3D CL is deduced using sweep angle.

    Contribution of high-lift devices is modelled according to their geometry (span and chord ratio)
    and their deflection angles.
    """

    def initialize(self):
        self.options.declare(
            "use_xfoil",
            default=True,
            types=bool,
            desc="If True, maximum 2D CL without high-lift "
            "aerodynamics:aircraft:landing:CL_max_clean_2D is computed using XFOIL.\n"
            "If False, aerodynamics:aircraft:landing:CL_max_clean_2D must be provided "
            "as input (but process is faster)",
        )
        self.options.declare(
            "xfoil_alpha_min",
            default=0.0,
            types=float,
            desc="Used if use_xfoil is True. Sets the minimum alpha that is explored "
            "to find maximum 2D CL without high-lift",
        )
        self.options.declare(
            "xfoil_alpha_max",
            default=30.0,
            types=float,
            desc="Used if use_xfoil is True. Sets the maximum alpha that is explored "
            "to find maximum 2D CL without high-lift",
        )
        self.options.declare(
            "xfoil_iter_limit",
            default=500,
            types=int,
            desc="Maximum number of iterations for a XFOIL run.",
        )
        self.options.declare(
            OPTION_XFOIL_EXE_PATH,
            default="",
            types=str,
            allow_none=True,
            desc="The path to the XFOIL executable. Needed for non-Windows OS.",
        )

    def setup(self):
        self.add_subsystem(
            "mach_reynolds",
            RegisterSubmodel.get_submodel(SERVICE_LANDING_MACH_REYNOLDS),
            promotes=["*"],
        )

        if self.options["use_xfoil"]:
            start = self.options["xfoil_alpha_min"]
            end = self.options["xfoil_alpha_max"]
            iter_limit = self.options["xfoil_iter_limit"]
            xfoil_options = {
                OPTION_ALPHA_START: start,
                OPTION_ALPHA_END: end,
                OPTION_ITER_LIMIT: iter_limit,
                OPTION_XFOIL_EXE_PATH: self.options[OPTION_XFOIL_EXE_PATH],
            }
            self.add_subsystem(
                "xfoil_run",
                RegisterSubmodel.get_submodel(SERVICE_XFOIL, xfoil_options),
                promotes=["data:geometry:wing:thickness_ratio"],
            )

        self.add_subsystem(
            "CL_2D_to_3D",
            RegisterSubmodel.get_submodel(SERVICE_LANDING_MAX_CL_CLEAN),
            promotes=["*"],
        )

        landing_flag_option = {"landing_flag": True}
        self.add_subsystem(
            "delta_cl_landing",
            RegisterSubmodel.get_submodel(SERVICE_HIGH_LIFT, landing_flag_option),
            promotes=["*"],
        )

        self.add_subsystem(
            "compute_max_cl_landing",
            RegisterSubmodel.get_submodel(SERVICE_LANDING_MAX_CL),
            promotes=["*"],
        )

        if self.options["use_xfoil"]:
            self.connect("data:aerodynamics:aircraft:landing:mach", "xfoil_run.xfoil:mach")
            self.connect("data:aerodynamics:wing:landing:reynolds", "xfoil_run.xfoil:reynolds")
            self.connect(
                "xfoil_run.xfoil:CL_max_2D", "data:aerodynamics:aircraft:landing:CL_max_clean_2D"
            )


@RegisterSubmodel(
    SERVICE_LANDING_MACH_REYNOLDS, "fastoad.submodel.aerodynamics.landing.mach_reynolds.legacy"
)
class ComputeMachReynolds(om.ExplicitComponent):
    """
    Mach and Reynolds computation
    """

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")
        self.add_output("data:aerodynamics:aircraft:landing:mach")
        self.add_output("data:aerodynamics:wing:landing:reynolds")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        speed = inputs["data:TLAR:approach_speed"]

        atm = Atmosphere(0.0, 15.0)
        atm.true_airspeed = speed
        reynolds = atm.unitary_reynolds * l0_wing

        outputs["data:aerodynamics:aircraft:landing:mach"] = atm.mach
        outputs["data:aerodynamics:wing:landing:reynolds"] = reynolds


@RegisterSubmodel(
    SERVICE_LANDING_MAX_CL_CLEAN, "fastoad.submodel.aerodynamics.landing.max_CL_clean.legacy"
)
class Compute3DMaxCL(om.ExplicitComponent):
    """
    Computes 3D max CL from 2D CL (XFOIL-computed) and sweep angle
    """

    def setup(self):
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="rad")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max_clean_2D", val=np.nan)

        self.add_output("data:aerodynamics:aircraft:landing:CL_max_clean")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        cl_max_2d = inputs["data:aerodynamics:aircraft:landing:CL_max_clean_2D"]
        outputs["data:aerodynamics:aircraft:landing:CL_max_clean"] = (
            cl_max_2d * 0.9 * np.cos(sweep_25)
        )



"""
Estimation of vertical tail distance
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


class ComputeVTDistance(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Vertical tail distance estimation"""

    def setup(self):
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("settings:geometry:vertical_tail:position_ratio_on_fuselage", val=0.88)

        self.add_output("data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25",
            ["data:geometry:fuselage:length", "data:geometry:wing:MAC:at25percent:x"],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        fus_length = inputs["data:geometry:fuselage:length"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        vtp_aero_center_ratio = inputs["settings:geometry:vertical_tail:position_ratio_on_fuselage"]

        lp_vt = vtp_aero_center_ratio * fus_length - fa_length

        outputs["data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25"] = lp_vt



"""
Estimation of vertical tail lift coefficient
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np
import openmdao.api as om


# TODO: This belongs more to aerodynamics than geometry


class ComputeVTClalpha(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Vertical tail lift coefficient estimation"""

    def setup(self):
        self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:geometry:has_T_tail", val=np.nan)
        self.add_input("data:geometry:vertical_tail:aspect_ratio", val=np.nan)
        self.add_input("data:geometry:vertical_tail:sweep_25", val=np.nan, units="deg")

        self.add_output("data:aerodynamics:vertical_tail:cruise:CL_alpha", units="1/rad")

    def setup_partials(self):
        self.declare_partials("data:aerodynamics:vertical_tail:cruise:CL_alpha", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        tail_type = np.round(inputs["data:geometry:has_T_tail"])
        cruise_mach = inputs["data:TLAR:cruise_mach"]
        sweep_25_vt = inputs["data:geometry:vertical_tail:sweep_25"]
        k_ar_effective = 2.9 if tail_type == 1 else 1.55
        lambda_vt = inputs["data:geometry:vertical_tail:aspect_ratio"] * k_ar_effective

        beta = math.sqrt(1 - cruise_mach**2)
        cl_alpha_vt = (
            0.8
            * 2
            * math.pi
            * lambda_vt
            / (
                2
                + math.sqrt(
                    4
                    + lambda_vt**2
                    * beta**2
                    / 0.95**2
                    * (1 + (math.tan(sweep_25_vt / 180.0 * math.pi)) ** 2 / beta**2)
                )
            )
        )

        outputs["data:aerodynamics:vertical_tail:cruise:CL_alpha"] = cl_alpha_vt



"""Computation of max CL in landing conditions."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.module_management.service_registry import RegisterSubmodel
from openmdao.core.explicitcomponent import ExplicitComponent

from ..constants import SERVICE_LANDING_MAX_CL


@RegisterSubmodel(SERVICE_LANDING_MAX_CL, "fastoad.submodel.aerodynamics.landing.max_CL.legacy")
class ComputeMaxClLanding(ExplicitComponent):
    """Computation of max CL in landing conditions."""

    def setup(self):
        self.add_input("data:aerodynamics:aircraft:landing:CL_max_clean", val=np.nan)
        self.add_input("data:aerodynamics:high_lift_devices:landing:CL", val=np.nan)
        self.add_input(
            "tuning:aerodynamics:aircraft:landing:CL_max:landing_gear_effect:k",
            val=np.nan,
        )
        self.add_output("data:aerodynamics:aircraft:landing:CL_max")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        cl_max_clean = inputs["data:aerodynamics:aircraft:landing:CL_max_clean"]
        cl_max_landing = cl_max_clean + inputs["data:aerodynamics:high_lift_devices:landing:CL"]
        cl_max_landing = (
            cl_max_landing
            * inputs["tuning:aerodynamics:aircraft:landing:CL_max:landing_gear_effect:k"]
        )

        outputs["data:aerodynamics:aircraft:landing:CL_max"] = cl_max_landing



"""Compressibility drag computation."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_CD_COMPRESSIBILITY


@RegisterSubmodel(
    SERVICE_CD_COMPRESSIBILITY,
    "fastoad.submodel.aerodynamics.CD.compressibility.legacy",
)
class CdCompressibility(om.ExplicitComponent):
    """
    Computation of drag increment due to compressibility effects.

    Formula from §4.2.4 of :cite:`supaero:2014`. This formula can be used for aircraft
    before year 2000.

    Earlier aircraft have more optimized wing profiles that are expected to limit the
    compressibility drag below 2 drag counts. Until a better model can be provided, the
    variable `tuning:aerodynamics:aircraft:cruise:CD:compressibility:characteristic_mach_increment`
    allows to move the characteristic Mach number, thus moving the CD divergence to higher
    Mach numbers.
    """

    def setup(self):
        self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:aerodynamics:aircraft:cruise:CL", shape_by_conn=True, val=np.nan)
        self.add_input("data:geometry:wing:sweep_25", units="deg", val=np.nan)
        self.add_input("data:geometry:wing:thickness_ratio", val=np.nan)
        self.add_input("tuning:aerodynamics:aircraft:cruise:CD:compressibility:max_value", val=0.5)
        self.add_input(
            "tuning:aerodynamics:aircraft:cruise:CD:compressibility:characteristic_mach_increment",
            val=0.0,
        )
        self.add_output(
            "data:aerodynamics:aircraft:cruise:CD:compressibility",
            copy_shape="data:aerodynamics:aircraft:cruise:CL",
        )

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        cl = inputs["data:aerodynamics:aircraft:cruise:CL"]
        m = inputs["data:TLAR:cruise_mach"]
        max_cd_comp = inputs["tuning:aerodynamics:aircraft:cruise:CD:compressibility:max_value"]
        delta_m_charac_0 = inputs[
            "tuning:aerodynamics:aircraft:cruise:CD:compressibility:characteristic_mach_increment"
        ]
        sweep_angle = inputs["data:geometry:wing:sweep_25"]
        thickness_ratio = inputs["data:geometry:wing:thickness_ratio"]

        # Computation of characteristic Mach for 28° sweep and 0.12 of relative thickness
        m_charac_comp_0 = (
            -0.5 * np.maximum(0.35, cl) ** 2
            + 0.35 * np.maximum(0.35, cl)
            + 0.765
            + delta_m_charac_0
        )

        # Computation of characteristic Mach for actual sweep angle and relative thickness
        m_charac_comp = (
            m_charac_comp_0 * np.cos(np.radians(28)) + 0.12 - thickness_ratio
        ) / np.cos(np.radians(sweep_angle))

        cd_comp = np.minimum(max_cd_comp, 0.002 * np.exp(42.58 * (m - m_charac_comp)))

        outputs["data:aerodynamics:aircraft:cruise:CD:compressibility"] = cd_comp



"""
Computation of load cases
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel
from stdatm import Atmosphere

from .constants import SERVICE_GUST_LOADS


@RegisterSubmodel(SERVICE_GUST_LOADS, "fastoad.submodel.gust_loads")
class Loads(om.ExplicitComponent):
    """
    Computes gust load cases

    Load case 1: with wings with almost no fuel
    Load case 2: at maximum take-off weight

    Based on formulas in :cite:`supaero:2014`, §6.3

    """

    def initialize(self):
        self.options.declare(
            "fuel_load_alleviation",
            types=bool,
            default=True,
            desc="If False this simulates a dry wing,"
            "i.e. the sizing load 2 does not take into account the fuel weight.",
        )

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:weight:aircraft:MZFW", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:cruise:CL_alpha", val=np.nan, units="1/rad")
        self.add_input("data:load_case:lc1:U_gust", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc1:altitude", val=np.nan, units="ft")
        self.add_input("data:load_case:lc1:Vc_EAS", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc2:U_gust", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc2:altitude", val=np.nan, units="ft")
        self.add_input("data:load_case:lc2:Vc_EAS", val=np.nan, units="m/s")
        self.add_input("data:load_case:manoeuvre_load_factor", val=2.5)
        self.add_input("data:load_case:gust_intensity", val=1.0)

        self.add_output("data:mission:sizing:cs25:load_factor_1")
        self.add_output("data:mission:sizing:cs25:load_factor_2")
        self.add_output("data:mission:sizing:cs25:sizing_load_1", units="kg")
        self.add_output("data:mission:sizing:cs25:sizing_load_2", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    # pylint: disable=too-many-locals
    # pylint: disable=invalid-name
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        sea_level_density = Atmosphere(0).density
        wing_area = inputs["data:geometry:wing:area"]
        span = inputs["data:geometry:wing:span"]
        mzfw = inputs["data:weight:aircraft:MZFW"]
        mfw = inputs["data:weight:aircraft:MFW"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        cl_alpha = inputs["data:aerodynamics:aircraft:cruise:CL_alpha"]
        u_gust1 = inputs["data:load_case:lc1:U_gust"]
        alt_1 = inputs["data:load_case:lc1:altitude"]
        vc_eas1 = inputs["data:load_case:lc1:Vc_EAS"]
        u_gust2 = inputs["data:load_case:lc2:U_gust"]
        alt_2 = inputs["data:load_case:lc2:altitude"]
        vc_eas2 = inputs["data:load_case:lc2:Vc_EAS"]
        n_manoeuvre = inputs["data:load_case:manoeuvre_load_factor"]
        gust_intensity = inputs["data:load_case:gust_intensity"]

        # calculation of mean geometric chord
        chord_geom = wing_area / span

        # load case #1
        m1 = 1.05 * mzfw
        n_gust_1 = self.__n_gust(
            m1,
            wing_area,
            Atmosphere(alt_1).density,
            sea_level_density,
            chord_geom,
            vc_eas1,
            cl_alpha,
            u_gust1,
        )
        n1 = 1.5 * max(n_manoeuvre, n_gust_1 * gust_intensity)
        n1m1 = n1 * m1

        # load case #2
        n_gust_2 = self.__n_gust(
            mtow,
            wing_area,
            Atmosphere(alt_2).density,
            sea_level_density,
            chord_geom,
            vc_eas2,
            cl_alpha,
            u_gust2,
        )
        n2 = 1.5 * max(n_manoeuvre, n_gust_2 * gust_intensity)

        if not self.options["fuel_load_alleviation"]:
            n2m2 = n2 * mtow
        else:
            mcv = min(0.8 * mfw, mtow - mzfw)
            n2m2 = n2 * (mtow - 0.55 * mcv)

        outputs["data:mission:sizing:cs25:load_factor_1"] = n1
        outputs["data:mission:sizing:cs25:load_factor_2"] = n2
        outputs["data:mission:sizing:cs25:sizing_load_1"] = n1m1
        outputs["data:mission:sizing:cs25:sizing_load_2"] = n2m2

    @staticmethod
    def __n_gust(mass, wing_area, rho, sea_level_density, chord_geom, vc_eas, cl_alpha, u_gust):
        """
        Computes a reference gust load.

        :param mass:
        :param wing_area:
        :param rho:
        :param sea_level_density:
        :param chord_geom:
        :param vc_eas: Vc (Equivalent AirSpeed)
        :param cl_alpha:
        :param u_gust:
        :return:
        """
        mu_g = 2 * mass / rho / wing_area / chord_geom / cl_alpha
        k_g = 0.88 * mu_g / (5.3 + mu_g)  # attenuation factor
        n_gust = 1 + (sea_level_density / 2 / 9.81) * k_g * u_gust * (
            vc_eas * cl_alpha / (mass / wing_area)
        )

        return n_gust



"""
Computation of Oswald coefficient
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_INDUCED_DRAG_COEFFICIENT, SERVICE_OSWALD_COEFFICIENT


@RegisterSubmodel(
    SERVICE_INDUCED_DRAG_COEFFICIENT,
    "fastoad.submodel.aerodynamics.induced_drag_coefficient.legacy",
)
class InducedDragCoefficient(om.ExplicitComponent):
    """Computes the coefficient that should be multiplied by CL**2 to get induced drag."""

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")

        if self.options["low_speed_aero"]:
            self.add_input("data:aerodynamics:aircraft:low_speed:oswald_coefficient", val=np.nan)
            self.add_output("data:aerodynamics:aircraft:low_speed:induced_drag_coefficient")
        else:
            self.add_input("data:aerodynamics:aircraft:cruise:oswald_coefficient", val=np.nan)
            self.add_output("data:aerodynamics:aircraft:cruise:induced_drag_coefficient")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        wing_area = inputs["data:geometry:wing:area"]
        span = inputs["data:geometry:wing:span"] / np.cos(5.0 / 180 * np.pi)
        aspect_ratio = span**2 / wing_area

        if self.options["low_speed_aero"]:
            coef_e = inputs["data:aerodynamics:aircraft:low_speed:oswald_coefficient"]
        else:
            coef_e = inputs["data:aerodynamics:aircraft:cruise:oswald_coefficient"]

        coef_k = 1.0 / (np.pi * aspect_ratio * coef_e)

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:aircraft:low_speed:induced_drag_coefficient"] = coef_k
        else:
            outputs["data:aerodynamics:aircraft:cruise:induced_drag_coefficient"] = coef_k


@RegisterSubmodel(
    SERVICE_OSWALD_COEFFICIENT,
    "fastoad.submodel.aerodynamics.oswald_coefficient.legacy",
)
class OswaldCoefficient(om.ExplicitComponent):
    """Computes Oswald efficiency number"""

    # TODO: Document equations. Cite sources (M. Nita and D. Scholz)

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="deg")

        if self.options["low_speed_aero"]:
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:aircraft:low_speed:oswald_coefficient")
        else:
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_output("data:aerodynamics:aircraft:cruise:oswald_coefficient")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        wing_area = inputs["data:geometry:wing:area"]
        span = inputs["data:geometry:wing:span"] / np.cos(5.0 / 180 * np.pi)
        height_fus = inputs["data:geometry:fuselage:maximum_height"]
        width_fus = inputs["data:geometry:fuselage:maximum_width"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        l4_wing = inputs["data:geometry:wing:tip:chord"]
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        if self.options["low_speed_aero"]:
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]
        else:
            mach = inputs["data:TLAR:cruise_mach"]

        aspect_ratio = span**2 / wing_area
        df = np.sqrt(width_fus * height_fus)
        lamda = l4_wing / l2_wing
        delta_lamda = -0.357 + 0.45 * np.exp(0.0375 * sweep_25 / 180.0 * np.pi)
        lamda = lamda - delta_lamda
        f_lamda = 0.0524 * lamda**4 - 0.15 * lamda**3 + 0.1659 * lamda**2 - 0.0706 * lamda + 0.0119
        e_theory = 1 / (1 + f_lamda * aspect_ratio)

        if mach <= 0.4:
            ke_m = 1.0
        else:
            ke_m = -0.001521 * ((mach - 0.05) / 0.3 - 1) ** 10.82 + 1

        ke_f = 1 - 2 * (df / span) ** 2
        coef_e = e_theory * ke_f * ke_m * 0.9

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:aircraft:low_speed:oswald_coefficient"] = coef_e
        else:
            outputs["data:aerodynamics:aircraft:cruise:oswald_coefficient"] = coef_e



"""
Computation of wing area following aerodynamic constraints
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel
from scipy.constants import g
from stdatm import Atmosphere

from fastoad_cs25.models.loops.constants import (
    SERVICE_WING_AREA_CONSTRAINT_AERO,
    SERVICE_WING_AREA_LOOP_AERO,
)


@RegisterSubmodel(
    SERVICE_WING_AREA_LOOP_AERO, "fastoad.submodel.loops.wing.area.update.aero.legacy"
)
class UpdateWingAreaAero(om.ExplicitComponent):
    """Computes wing area for having enough lift at required approach speed."""

    def setup(self):
        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")
        self.add_input("data:weight:aircraft:MLW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max", val=np.nan)

        # My plan is to not promote this variable and connect it by hand so that it foes not
        # appear in the output file as is but only as a constraints in the other component. We
        # could however give it a proper name.
        self.add_output("wing_area:aero", val=100.0, units="m**2")

        self.declare_partials(of="*", wrt="*", method="exact")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        approach_speed = inputs["data:TLAR:approach_speed"]
        mlw = inputs["data:weight:aircraft:MLW"]
        max_cl = inputs["data:aerodynamics:aircraft:landing:CL_max"]

        rho_sl = Atmosphere(0.0).density
        stall_speed = approach_speed / 1.23

        wing_area_approach = mlw * g / (0.5 * rho_sl * stall_speed**2.0 * max_cl)

        outputs["wing_area:aero"] = wing_area_approach

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        approach_speed = inputs["data:TLAR:approach_speed"]
        mlw = inputs["data:weight:aircraft:MLW"]
        max_cl = inputs["data:aerodynamics:aircraft:landing:CL_max"]

        rho_sl = Atmosphere(0.0).density
        stall_speed = approach_speed / 1.23

        partials["wing_area:aero", "data:TLAR:approach_speed"] = (
            -2.0 * (mlw * g) / (0.5 * rho_sl * stall_speed**3.0 * max_cl) / 1.23
        )
        partials["wing_area:aero", "data:weight:aircraft:MLW"] = g / (
            0.5 * rho_sl * stall_speed**2.0 * max_cl
        )
        partials["wing_area:aero", "data:aerodynamics:aircraft:landing:CL_max"] = -(mlw * g) / (
            0.5 * rho_sl * stall_speed**2.0 * max_cl**2.0
        )


@RegisterSubmodel(
    SERVICE_WING_AREA_CONSTRAINT_AERO,
    "fastoad.submodel.loops.wing.area.constraint.aero.legacy",
)
class WingAreaConstraintsAero(om.ExplicitComponent):
    def setup(self):
        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")
        self.add_input("data:weight:aircraft:MLW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max", val=np.nan)
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")

        self.add_output("data:aerodynamics:aircraft:landing:additional_CL_capacity")

        self.declare_partials(of="*", wrt="*", method="exact")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        approach_speed = inputs["data:TLAR:approach_speed"]
        mlw = inputs["data:weight:aircraft:MLW"]
        max_cl = inputs["data:aerodynamics:aircraft:landing:CL_max"]
        wing_area = inputs["data:geometry:wing:area"]

        rho_sl = Atmosphere(0.0).density
        stall_speed = approach_speed / 1.23

        outputs["data:aerodynamics:aircraft:landing:additional_CL_capacity"] = max_cl - mlw * g / (
            0.5 * rho_sl * stall_speed**2 * wing_area
        )

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        approach_speed = inputs["data:TLAR:approach_speed"]
        mlw = inputs["data:weight:aircraft:MLW"]
        wing_area = inputs["data:geometry:wing:area"]

        rho_sl = Atmosphere(0.0).density
        stall_speed = approach_speed / 1.23

        partials[
            "data:aerodynamics:aircraft:landing:additional_CL_capacity",
            "data:TLAR:approach_speed",
        ] = 2.0 * (mlw * g) / (0.5 * rho_sl * stall_speed**3.0 * wing_area) / 1.23
        partials[
            "data:aerodynamics:aircraft:landing:additional_CL_capacity",
            "data:aerodynamics:aircraft:landing:CL_max",
        ] = 1.0
        partials[
            "data:aerodynamics:aircraft:landing:additional_CL_capacity",
            "data:weight:aircraft:MLW",
        ] = -g / (0.5 * rho_sl * stall_speed**2 * wing_area)
        partials[
            "data:aerodynamics:aircraft:landing:additional_CL_capacity",
            "data:geometry:wing:area",
        ] = (mlw * g) / (0.5 * rho_sl * stall_speed**2 * wing_area**2.0)



"""
Estimation of fixed operational systems weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_FIXED_OPERATIONAL_SYSTEMS_MASS


@RegisterSubmodel(
    SERVICE_FIXED_OPERATIONAL_SYSTEMS_MASS,
    "fastoad.submodel.weight.mass.systems.fixed_operational.legacy",
)
class FixedOperationalSystemsWeight(om.ExplicitComponent):
    """
    Weight estimation for fixed operational systems (weather radar, flight recorder, ...)

    Based on formulas in :cite:`supaero:2014`, mass contribution C5
    """

    def setup(self):
        self.add_input("data:geometry:fuselage:front_length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:rear_length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:seats:economical:count_by_row", val=np.nan)
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:containers:count_by_row", val=np.nan)
        self.add_input("tuning:weight:systems:operational:mass:k", val=1.0)
        self.add_input("tuning:weight:systems:operational:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:systems:operational:radar:mass", units="kg")
        self.add_output("data:weight:systems:operational:cargo_hold:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    # pylint: disable=too-many-locals
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        front_section_length = inputs["data:geometry:fuselage:front_length"]
        aft_section_length = inputs["data:geometry:fuselage:rear_length"]
        fuselage_length = inputs["data:geometry:fuselage:length"]
        front_seat_number_eco = inputs["data:geometry:cabin:seats:economical:count_by_row"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        side_by_side_container_number = inputs["data:geometry:cabin:containers:count_by_row"]
        k_c5 = inputs["tuning:weight:systems:operational:mass:k"]
        offset_c5 = inputs["tuning:weight:systems:operational:mass:offset"]

        cylindrical_section_length = fuselage_length - front_section_length - aft_section_length

        cargo_compartment_length = (
            cylindrical_section_length + 0.864 * (front_seat_number_eco - 5) - 0.8 * l2_wing
        )

        # Radar
        temp_c51 = 100.0
        outputs["data:weight:systems:operational:radar:mass"] = k_c5 * temp_c51 + offset_c5

        # Cargo container systems
        temp_c52 = 23.4 * cargo_compartment_length * side_by_side_container_number
        outputs["data:weight:systems:operational:cargo_hold:mass"] = k_c5 * temp_c52 + offset_c5



"""
Estimation of horizontal tail chords and span
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

import openmdao.api as om


# TODO: is an OpenMDAO component required for this simple calculation ?
class ComputeHTChord(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Horizontal tail chords and span estimation"""

    def setup(self):
        self.add_input("data:geometry:horizontal_tail:aspect_ratio", val=np.nan)
        self.add_input("data:geometry:horizontal_tail:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:horizontal_tail:taper_ratio", val=np.nan)

        self.add_output("data:geometry:horizontal_tail:span", units="m")
        self.add_output("data:geometry:horizontal_tail:center:chord", units="m")
        self.add_output("data:geometry:horizontal_tail:tip:chord", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:horizontal_tail:span",
            ["data:geometry:horizontal_tail:area", "data:geometry:horizontal_tail:aspect_ratio"],
            method="fd",
        )
        self.declare_partials("data:geometry:horizontal_tail:center:chord", "*", method="fd")
        self.declare_partials("data:geometry:horizontal_tail:tip:chord", "*", method="fd")

    def compute(self, inputs, outputs):
        lambda_ht = inputs["data:geometry:horizontal_tail:aspect_ratio"]
        s_h = inputs["data:geometry:horizontal_tail:area"]
        taper_ht = inputs["data:geometry:horizontal_tail:taper_ratio"]

        b_h = np.sqrt(max(lambda_ht * s_h, 0.1))
        root_chord = s_h * 2 / (1 + taper_ht) / b_h
        tip_chord = root_chord * taper_ht

        outputs["data:geometry:horizontal_tail:span"] = b_h
        outputs["data:geometry:horizontal_tail:center:chord"] = root_chord
        outputs["data:geometry:horizontal_tail:tip:chord"] = tip_chord



"""
Submodel for computing wing planform.
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad

from .compute_b_50 import ComputeB50
from .compute_l1_l4 import ComputeL1AndL4Wing
from .compute_l2_l3 import ComputeL2AndL3Wing
from .compute_mac_wing import ComputeMACWing
from .compute_sweep_wing import ComputeSweepWing
from .compute_x_wing import ComputeXWing
from .compute_y_wing import ComputeYWing
from ..constants import SERVICE_WING_GEOMETRY_PLANFORM


@oad.RegisterSubmodel(
    SERVICE_WING_GEOMETRY_PLANFORM, "fastoad.submodel.geometry.wing.planform.legacy"
)
class ComputeWingGeometry(
    oad.CycleGroup,
    default_linear_options={"iprint": 0},
    default_nonlinear_options={"maxiter": 50, "iprint": 0},
):
    """Computation of wing planform"""

    def setup(self):
        super().setup()
        self.add_subsystem("y_wing", ComputeYWing(), promotes=["*"])
        self.add_subsystem("l14_wing", ComputeL1AndL4Wing(), promotes=["*"])
        self.add_subsystem("l2l3_wing", ComputeL2AndL3Wing(), promotes=["*"])
        self.add_subsystem("x_wing", ComputeXWing(), promotes=["*"])
        self.add_subsystem("mac_wing", ComputeMACWing(), promotes=["*"])
        self.add_subsystem("b50_wing", ComputeB50(), promotes=["*"])
        self.add_subsystem("sweep_wing", ComputeSweepWing(), promotes=["*"])



"""Computation of aerodynamic polar in low speed conditions."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterSubmodel

from .constants import (
    PolarType,
    SERVICE_ALPHA,
    SERVICE_CD0,
    SERVICE_CD_TRIM,
    SERVICE_CL_ALPHA,
    SERVICE_INDUCED_DRAG_COEFFICIENT,
    SERVICE_INITIALIZE_CL,
    SERVICE_OSWALD_COEFFICIENT,
    SERVICE_POLAR,
    SERVICE_REYNOLDS_COEFFICIENT,
)


@RegisterOpenMDAOSystem("fastoad.aerodynamics.lowspeed.legacy", domain=ModelDomain.AERODYNAMICS)
class AerodynamicsLowSpeed(om.Group):
    """
    Models for low speed aerodynamics
    """

    def setup(self):
        low_speed_option = {"low_speed_aero": True}

        ivc = om.IndepVarComp("data:aerodynamics:aircraft:takeoff:mach", val=0.2)
        self.add_subsystem("mach_low_speed", ivc, promotes=["*"])

        self.add_subsystem(
            "compute_oswald_coeff",
            RegisterSubmodel.get_submodel(SERVICE_OSWALD_COEFFICIENT, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "compute_induced_drag_coeff",
            RegisterSubmodel.get_submodel(SERVICE_INDUCED_DRAG_COEFFICIENT, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "comp_re",
            RegisterSubmodel.get_submodel(SERVICE_REYNOLDS_COEFFICIENT, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "initialize_cl",
            RegisterSubmodel.get_submodel(SERVICE_INITIALIZE_CL, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_wing", RegisterSubmodel.get_submodel(SERVICE_CD0, low_speed_option), promotes=["*"]
        )
        self.add_subsystem(
            "cd_trim",
            RegisterSubmodel.get_submodel(SERVICE_CD_TRIM, low_speed_option),
            promotes=["*"],
        )
        polar_type_option = {"polar_type": PolarType.LOW_SPEED}
        self.add_subsystem(
            "get_polar",
            RegisterSubmodel.get_submodel(SERVICE_POLAR, polar_type_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "compute_CLalpha",
            RegisterSubmodel.get_submodel(SERVICE_CL_ALPHA, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "compute_alpha",
            RegisterSubmodel.get_submodel(SERVICE_ALPHA, low_speed_option),
            promotes=["*"],
        )



"""Computation of form drag for nacelles and pylons."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .utils.friction_drag import get_flat_plate_friction_drag_coefficient
from ..constants import SERVICE_CD0_NACELLES_PYLONS


@RegisterSubmodel(
    SERVICE_CD0_NACELLES_PYLONS,
    "fastoad.submodel.aerodynamics.CD0.nacelles_pylons.legacy",
)
class Cd0NacellesAndPylons(om.ExplicitComponent):
    """Computation of form drag for nacelles and pylons."""

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        if self.options["low_speed_aero"]:
            self.add_input("data:aerodynamics:wing:low_speed:reynolds", val=np.nan)
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:nacelles:low_speed:CD0")
            self.add_output("data:aerodynamics:pylons:low_speed:CD0")
        else:
            self.add_input("data:aerodynamics:wing:cruise:reynolds", val=np.nan)
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_output("data:aerodynamics:nacelles:cruise:CD0")
            self.add_output("data:aerodynamics:pylons:cruise:CD0")

        self.add_input("data:geometry:propulsion:pylon:length", val=np.nan, units="m")
        self.add_input("data:geometry:propulsion:nacelle:length", val=np.nan, units="m")
        self.add_input("data:geometry:propulsion:pylon:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:propulsion:nacelle:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)
        self.add_input("data:geometry:propulsion:fan:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        n_engines = inputs["data:geometry:propulsion:engine:count"]
        wing_area = inputs["data:geometry:wing:area"]
        if self.options["low_speed_aero"]:
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]
            reynolds = inputs["data:aerodynamics:wing:low_speed:reynolds"]
        else:
            mach = inputs["data:TLAR:cruise_mach"]
            reynolds = inputs["data:aerodynamics:wing:cruise:reynolds"]

        cd0_pylon = self._compute_cd0_for_pylons(inputs, n_engines, wing_area, mach, reynolds)
        cd0_nac = self._compute_cd0_for_nacelles(inputs, n_engines, wing_area, mach, reynolds)

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:pylons:low_speed:CD0"] = cd0_pylon
            outputs["data:aerodynamics:nacelles:low_speed:CD0"] = cd0_nac
        else:
            outputs["data:aerodynamics:pylons:cruise:CD0"] = cd0_pylon
            outputs["data:aerodynamics:nacelles:cruise:CD0"] = cd0_nac

    @staticmethod
    def _compute_cd0_for_pylons(inputs, n_engines, wing_area, mach, reynolds):
        pylon_length = inputs["data:geometry:propulsion:pylon:length"]
        wet_area_pylon = inputs["data:geometry:propulsion:pylon:wetted_area"]

        cf_pylon = get_flat_plate_friction_drag_coefficient(pylon_length, mach, reynolds)
        el_pylon = 0.06
        ke_cd0_pylon = 4.688 * el_pylon**2 + 3.146 * el_pylon
        cd0_pylon = n_engines * (1 + ke_cd0_pylon) * cf_pylon * wet_area_pylon / wing_area
        return cd0_pylon

    @staticmethod
    def _compute_cd0_for_nacelles(inputs, n_engines, wing_area, mach, reynolds):
        nac_length = inputs["data:geometry:propulsion:nacelle:length"]
        wet_area_nac = inputs["data:geometry:propulsion:nacelle:wetted_area"]
        fan_length = inputs["data:geometry:propulsion:fan:length"]

        cf_nac = get_flat_plate_friction_drag_coefficient(nac_length, mach, reynolds)
        e_fan = 0.22
        kn_cd0_nac = 1 + 0.05 + 5.8 * e_fan / fan_length
        cd0_int_nac = 0.0002
        cd0_nac = n_engines * (kn_cd0_nac * cf_nac * wet_area_nac / wing_area + cd0_int_nac)
        return cd0_nac



"""
Estimation of cargo configuration weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_CARGO_CONFIGURATION_MASS


@RegisterSubmodel(
    SERVICE_CARGO_CONFIGURATION_MASS, "service.mass.furniture.cargo_configuration.legacy"
)
class CargoConfigurationWeight(om.ExplicitComponent):
    """
    Weight estimation for equipments for freight transport (applies only for freighters)

    Based on :cite:`supaero:2014`, mass contribution D1
    """

    def setup(self):
        self.add_input("data:geometry:cabin:NPAX1", val=np.nan)
        self.add_input("data:geometry:cabin:containers:count", val=np.nan)
        self.add_input("data:geometry:cabin:pallet_count", val=np.nan)
        self.add_input("data:geometry:cabin:seats:economical:count_by_row", val=np.nan)
        self.add_input("tuning:weight:furniture:cargo_configuration:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:furniture:cargo_configuration:mass:offset", val=0.0, units="kg"
        )

        self.add_output("data:weight:furniture:cargo_configuration:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        npax1 = inputs["data:geometry:cabin:NPAX1"]
        side_by_side_eco_seat_count = inputs["data:geometry:cabin:seats:economical:count_by_row"]
        container_count = inputs["data:geometry:cabin:containers:count"]
        pallet_number = inputs["data:geometry:cabin:pallet_count"]
        k_d1 = inputs["tuning:weight:furniture:cargo_configuration:mass:k"]
        offset_d1 = inputs["tuning:weight:furniture:cargo_configuration:mass:offset"]

        if side_by_side_eco_seat_count <= 6.0:
            temp_d1 = 0.351 * (npax1 - 38)
        else:
            temp_d1 = 85 * container_count + 110 * pallet_number
        outputs["data:weight:furniture:cargo_configuration:mass"] = k_d1 * temp_d1 + offset_d1



"""
Estimation of fuselage weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_FUSELAGE_MASS


@RegisterSubmodel(SERVICE_FUSELAGE_MASS, "fastoad.submodel.weight.mass.airframe.fuselage.legacy")
class FuselageWeight(om.ExplicitComponent):
    """
    Fuselage weight estimation

    Based on a statistical analysis. See :cite:`supaero:2014`, mass contribution A2
    """

    def setup(self):
        self.add_input("data:geometry:fuselage:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:mission:sizing:cs25:sizing_load_1", val=np.nan, units="kg")
        self.add_input("tuning:weight:airframe:fuselage:mass:k", val=1.0)
        self.add_input("tuning:weight:airframe:fuselage:mass:offset", val=0.0, units="kg")
        self.add_input("settings:weight:airframe:fuselage:mass:k_lg", val=1.05)
        self.add_input("settings:weight:airframe:fuselage:mass:k_fus", val=1.0)

        self.add_output("data:weight:airframe:fuselage:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        fuselage_wet_area = inputs["data:geometry:fuselage:wetted_area"]
        width_max = inputs["data:geometry:fuselage:maximum_width"]
        height_max = inputs["data:geometry:fuselage:maximum_height"]
        n1m1 = inputs["data:mission:sizing:cs25:sizing_load_1"]
        k_a2 = inputs["tuning:weight:airframe:fuselage:mass:k"]
        offset_a2 = inputs["tuning:weight:airframe:fuselage:mass:offset"]
        k_lg = inputs["settings:weight:airframe:fuselage:mass:k_lg"]
        k_fus = inputs["settings:weight:airframe:fuselage:mass:k_fus"]

        temp_a2 = (
            fuselage_wet_area
            * (10 + 1.2 * np.sqrt(width_max * height_max) + 0.00019 * n1m1 / height_max**1.7)
            * k_lg
            * k_fus
        )
        outputs["data:weight:airframe:fuselage:mass"] = k_a2 * temp_a2 + offset_a2



"""
Estimation of life support systems weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.constants import RangeCategory
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_LIFE_SUPPORT_SYSTEMS_MASS


@RegisterSubmodel(
    SERVICE_LIFE_SUPPORT_SYSTEMS_MASS, "fastoad.submodel.weight.mass.systems.life_support.legacy"
)
class LifeSupportSystemsWeight(om.ExplicitComponent):
    """
    Weight estimation for life support systems

    This includes:

    - insulation
    - air conditioning / pressurization
    - de-icing
    - internal lighting system
    - seats and installation of crew
    - fixed oxygen
    - permanent security kits

    Based on formulas in :cite:`supaero:2014`, mass contribution C2
    """

    def setup(self):
        self.add_input("data:TLAR:range", val=np.nan, units="NM")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:sweep_0", val=np.nan, units="rad")
        self.add_input("data:geometry:propulsion:nacelle:diameter", val=np.nan, units="m")
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)
        self.add_input("data:geometry:cabin:NPAX1", val=np.nan)
        self.add_input("data:geometry:cabin:crew_count:technical", val=np.nan)
        self.add_input("data:geometry:cabin:crew_count:commercial", val=np.nan)
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:weight:propulsion:engine:mass", val=np.nan, units="kg")
        self.add_input("tuning:weight:systems:life_support:insulation:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:life_support:insulation:mass:offset", val=0.0, units="kg"
        )
        self.add_input("tuning:weight:systems:life_support:air_conditioning:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:life_support:air_conditioning:mass:offset", val=0.0, units="kg"
        )
        self.add_input("tuning:weight:systems:life_support:de-icing:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:life_support:de-icing:mass:offset", val=0.0, units="kg"
        )
        self.add_input("tuning:weight:systems:life_support:cabin_lighting:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:life_support:cabin_lighting:mass:offset", val=0.0, units="kg"
        )
        self.add_input(
            "tuning:weight:systems:life_support:seats_crew_accommodation:mass:k", val=1.0
        )
        self.add_input(
            "tuning:weight:systems:life_support:seats_crew_accommodation:mass:offset",
            val=0.0,
            units="kg",
        )
        self.add_input("tuning:weight:systems:life_support:oxygen:mass:k", val=1.0)
        self.add_input("tuning:weight:systems:life_support:oxygen:mass:offset", val=0.0, units="kg")
        self.add_input("tuning:weight:systems:life_support:safety_equipment:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:life_support:safety_equipment:mass:offset", val=0.0, units="kg"
        )

        self.add_output("data:weight:systems:life_support:insulation:mass", units="kg")
        self.add_output("data:weight:systems:life_support:air_conditioning:mass", units="kg")
        self.add_output("data:weight:systems:life_support:de-icing:mass", units="kg")
        self.add_output("data:weight:systems:life_support:cabin_lighting:mass", units="kg")
        self.add_output(
            "data:weight:systems:life_support:seats_crew_accommodation:mass", units="kg"
        )
        self.add_output("data:weight:systems:life_support:oxygen:mass", units="kg")
        self.add_output("data:weight:systems:life_support:safety_equipment:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    # pylint: disable=too-many-locals
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        tlar_range = inputs["data:TLAR:range"]
        width_max = inputs["data:geometry:fuselage:maximum_width"]
        height_max = inputs["data:geometry:fuselage:maximum_height"]
        cabin_length = inputs["data:geometry:cabin:length"]
        sweep_leading_edge = inputs["data:geometry:wing:sweep_0"]
        n_engines = inputs["data:geometry:propulsion:engine:count"]
        span = inputs["data:geometry:wing:span"]
        nacelle_diameter = inputs["data:geometry:propulsion:nacelle:diameter"]
        npax1 = inputs["data:geometry:cabin:NPAX1"]
        weight_engines = inputs["data:weight:propulsion:engine:mass"]
        cabin_crew = inputs["data:geometry:cabin:crew_count:commercial"]
        cockpit_crew = inputs["data:geometry:cabin:crew_count:technical"]
        k_c21 = inputs["tuning:weight:systems:life_support:insulation:mass:k"]
        offset_c21 = inputs["tuning:weight:systems:life_support:insulation:mass:offset"]
        k_c22 = inputs["tuning:weight:systems:life_support:air_conditioning:mass:k"]
        offset_c22 = inputs["tuning:weight:systems:life_support:air_conditioning:mass:offset"]
        k_c23 = inputs["tuning:weight:systems:life_support:de-icing:mass:k"]
        offset_c23 = inputs["tuning:weight:systems:life_support:de-icing:mass:offset"]
        k_c24 = inputs["tuning:weight:systems:life_support:cabin_lighting:mass:k"]
        offset_c24 = inputs["tuning:weight:systems:life_support:cabin_lighting:mass:offset"]
        k_c25 = inputs["tuning:weight:systems:life_support:seats_crew_accommodation:mass:k"]
        offset_c25 = inputs[
            "tuning:weight:systems:life_support:seats_crew_accommodation:mass:offset"
        ]
        k_c26 = inputs["tuning:weight:systems:life_support:oxygen:mass:k"]
        offset_c26 = inputs["tuning:weight:systems:life_support:oxygen:mass:offset"]
        k_c27 = inputs["tuning:weight:systems:life_support:safety_equipment:mass:k"]
        offset_c27 = inputs["tuning:weight:systems:life_support:safety_equipment:mass:offset"]

        fuselage_diameter = np.sqrt(width_max * height_max)

        # Mass of insulating system
        temp_c21 = 9.3 * fuselage_diameter * cabin_length
        outputs["data:weight:systems:life_support:insulation:mass"] = k_c21 * temp_c21 + offset_c21

        # Mass of air conditioning and pressurization system
        if tlar_range <= RangeCategory.MEDIUM.max():
            temp_c22 = (
                200
                + 27 * npax1**0.46
                + 7.2 * (n_engines**0.7) * (npax1**0.64)
                + npax1
                + 0.0029 * npax1**1.64
            )
        else:
            temp_c22 = (
                450
                + 51 * npax1**0.46
                + 7.2 * (n_engines**0.7) * (npax1**0.64)
                + npax1
                + 0.0029 * npax1**1.64
            )
        outputs["data:weight:systems:life_support:air_conditioning:mass"] = (
            k_c22 * temp_c22 + offset_c22
        )

        # Mass of de-icing system
        temp_c23 = (
            53
            + 9.5 * nacelle_diameter * n_engines
            + 1.9 * (span - width_max) / np.cos(sweep_leading_edge)
        )
        outputs["data:weight:systems:life_support:de-icing:mass"] = k_c23 * temp_c23 + offset_c23

        # Mass of internal lighting system
        temp_c24 = 1.4 * cabin_length * fuselage_diameter
        outputs["data:weight:systems:life_support:cabin_lighting:mass"] = (
            k_c24 * temp_c24 + offset_c24
        )

        # Mass of seats and installation system
        temp_c25 = 27 * cockpit_crew + 18 * cabin_crew
        outputs["data:weight:systems:life_support:seats_crew_accommodation:mass"] = (
            k_c25 * temp_c25 + offset_c25
        )

        # Mass of fixed oxygen
        temp_c26 = 80 + 1.3 * npax1
        outputs["data:weight:systems:life_support:oxygen:mass"] = k_c26 * temp_c26 + offset_c26

        # Mass of permanent security kits
        temp_c27 = 0.01 * weight_engines + 2.30 * npax1
        outputs["data:weight:systems:life_support:safety_equipment:mass"] = (
            k_c27 * temp_c27 + offset_c27
        )



"""Computation of mass of systems."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.module_management.service_registry import RegisterSubmodel
from openmdao import api as om

from .constants import (
    SERVICE_FIXED_OPERATIONAL_SYSTEMS_MASS,
    SERVICE_FLIGHT_KIT_MASS,
    SERVICE_LIFE_SUPPORT_SYSTEMS_MASS,
    SERVICE_NAVIGATION_SYSTEMS_MASS,
    SERVICE_POWER_SYSTEMS_MASS,
    SERVICE_TRANSMISSION_SYSTEMS_MASS,
)
from ..constants import SERVICE_SYSTEMS_MASS


@RegisterSubmodel(SERVICE_SYSTEMS_MASS, "fastoad.submodel.weight.mass.systems.legacy")
class SystemsWeight(om.Group):
    """
    Computes mass of systems.
    """

    def setup(self):
        self.add_subsystem(
            "power_systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_POWER_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "life_support_systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_LIFE_SUPPORT_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "navigation_systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_NAVIGATION_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "transmission_systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_TRANSMISSION_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "fixed_operational_systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_FIXED_OPERATIONAL_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "flight_kit_weight",
            RegisterSubmodel.get_submodel(SERVICE_FLIGHT_KIT_MASS),
            promotes=["*"],
        )

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:systems:mass",
            [
                "data:weight:systems:power:auxiliary_power_unit:mass",
                "data:weight:systems:power:electric_systems:mass",
                "data:weight:systems:power:hydraulic_systems:mass",
                "data:weight:systems:life_support:insulation:mass",
                "data:weight:systems:life_support:air_conditioning:mass",
                "data:weight:systems:life_support:de-icing:mass",
                "data:weight:systems:life_support:cabin_lighting:mass",
                "data:weight:systems:life_support:seats_crew_accommodation:mass",
                "data:weight:systems:life_support:oxygen:mass",
                "data:weight:systems:life_support:safety_equipment:mass",
                "data:weight:systems:navigation:mass",
                "data:weight:systems:transmission:mass",
                "data:weight:systems:operational:radar:mass",
                "data:weight:systems:operational:cargo_hold:mass",
                "data:weight:systems:flight_kit:mass",
            ],
            units="kg",
            desc="Mass of aircraft systems",
        )

        self.add_subsystem("systems_weight_sum", weight_sum, promotes=["*"])



"""Computes and aggregates CG ratios for load cases."""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from itertools import count

import numpy as np
from fastoad.module_management.exceptions import FastNoSubmodelFoundError
from fastoad.module_management.service_registry import RegisterSubmodel
from openmdao import api as om

from ...constants import SERVICE_LOAD_CASES_CG

SERVICE_LOAD_CASE_CG_PREFIX = "service.cg.load_case"


@RegisterSubmodel(SERVICE_LOAD_CASES_CG, "fastoad.submodel.weight.cg.load_cases.legacy")
class CGRatiosForLoadCases(om.Group):
    """Aggregation of CG ratios from load case calculations."""

    def setup(self):
        # We add in our group all the components for declared services that provide CG ratio
        # for specific load_cases
        for load_case_count in count():
            try:
                system = RegisterSubmodel.get_submodel(
                    f"{SERVICE_LOAD_CASE_CG_PREFIX}.{load_case_count + 1}"
                )
            except FastNoSubmodelFoundError:
                break
            self.add_subsystem(f"cg_ratio_lc{load_case_count + 1}", system, promotes_inputs=["*"])

        cg_ratio_aggregator = self.add_subsystem(
            "cg_ratio_aggregator",
            om.MuxComp(vec_size=load_case_count),
            promotes=["data:weight:aircraft:load_cases:CG:MAC_position"],
        )

        # This part aggregates all CG ratios values in one vector variable.
        cg_ratio_aggregator.add_var(
            "data:weight:aircraft:load_cases:CG:MAC_position", shape=(1,), axis=0
        )
        for i in range(load_case_count):
            self.connect(
                f"cg_ratio_lc{i + 1}.data:weight:aircraft:load_case_{i + 1}:CG:MAC_position",
                f"cg_ratio_aggregator.data:weight:aircraft:load_cases:CG:MAC_position_{i}",
            )

        self.add_subsystem("compute_max", MaxCGRatiosForLoadCases(), promotes=["*"])


class MaxCGRatiosForLoadCases(om.ExplicitComponent):
    """Maximum center of gravity ratio from load cases."""

    def setup(self):
        self.add_input(
            "data:weight:aircraft:load_cases:CG:MAC_position", val=np.nan, shape_by_conn=True
        )

        self.add_output("data:weight:aircraft:load_cases:CG:MAC_position:maximum")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        outputs["data:weight:aircraft:load_cases:CG:MAC_position:maximum"] = (
            np.nanmax(inputs["data:weight:aircraft:load_cases:CG:MAC_position"]),
        )



"""
Constants for rubber engine analytical models
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from stdatm import AtmosphereSI

RUBBER_ENGINE_DESCRIPTION = """
Parametric engine model as OpenMDAO component.

Implementation of E. Roux models for fuel consumption of low bypass ratio engines
For more information, see RubberEngine class in FAST-OAD developer documentation.
"""

# Atmosphere at limits of troposhere
ATM_SEA_LEVEL = AtmosphereSI(0)
ATM_TROPOPAUSE = AtmosphereSI(11000)

# Constants for computation of maximum thrust ---------------------------------
# (see E. Roux model definition in roux:2005)
A_MS = -2.74e-4
A_FM = 2.67e-4
B_MS = 1.91e-2
B_FM = -2.35e-2
C_MS = 1.21e-3
C_FM = -1.32e-3
D_MS = -8.48e-4
D_FM = 3.14e-4
E_MS = 8.96e-1
E_FM = 5.22e-1

ALPHA = [
    [1.79e-12, 4.29e-13, -5.24e-14, -4.51e-14, -4.57e-12],
    [1.17e-8, -8.80e-8, -5.25e-9, -3.19e-9, 5.52e-8],
    [-5.37e-13, -1.26e-12, 1.29e-14, 2.39e-14, 2.35e-12],
    [-3.18e-9, 2.76e-8, 1.97e-9, 1.17e-9, -2.26e-8],
]

BETA = [
    [1.70e-12, 1.51e-12, 1.48e-9, -7.59e-14, -1.07e-11],
    [-3.48e-9, -8.41e-8, 2.56e-5, -2.00e-8, -7.17e-8],
    [-3.89e-13, -2.05e-12, -9.28e-10, 1.30e-13, 5.39e-12],
    [1.77e-9, 2.62e-8, -8.87e-6, 6.66e-9, 4.43e-8],
]
# -----------------------------------------------------------------------------

# Constants for computation of SFC ratio --------------------------------------
MAX_SFC_RATIO_COEFF = 0.5
# -----------------------------------------------------------------------------



"""
Estimation of maximum center of gravity ratio
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


class ComputeMaxCGratio(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Maximum center of gravity ratio estimation"""

    def setup(self):
        self.add_input("data:weight:aircraft:empty:CG:MAC_position", val=np.nan)
        self.add_input("data:weight:aircraft:load_cases:CG:MAC_position:maximum", val=np.nan)
        self.add_input(
            "settings:weight:aircraft:CG:aft:MAC_position:margin",
            val=0.05,
            desc="Added margin for getting most aft CG position, "
            "as ratio of mean aerodynamic chord",
        )

        self.add_output("data:weight:aircraft:CG:aft:MAC_position")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        outputs["data:weight:aircraft:CG:aft:MAC_position"] = inputs[
            "settings:weight:aircraft:CG:aft:MAC_position:margin"
        ] + np.nanmax(
            [
                inputs["data:weight:aircraft:empty:CG:MAC_position"],
                inputs["data:weight:aircraft:load_cases:CG:MAC_position:maximum"],
            ]
        )



"""Computation of aerodynamic characteristics at takeoff."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import (
    RegisterOpenMDAOSystem,
    RegisterSubmodel,
)

from .constants import PolarType, SERVICE_HIGH_LIFT, SERVICE_POLAR


@RegisterOpenMDAOSystem("fastoad.aerodynamics.takeoff.legacy", domain=ModelDomain.AERODYNAMICS)
class AerodynamicsTakeoff(om.Group):
    """
    Computes aerodynamic characteristics at takeoff.

    - Computes CL and CD increments due to high-lift devices at takeoff.
    """

    def setup(self):
        landing_flag_option = {"landing_flag": False}
        self.add_subsystem(
            "delta_cl_cd",
            RegisterSubmodel.get_submodel(SERVICE_HIGH_LIFT, landing_flag_option),
            promotes=["*"],
        )

        polar_type_option = {"polar_type": PolarType.TAKEOFF}
        self.add_subsystem(
            "polar",
            RegisterSubmodel.get_submodel(SERVICE_POLAR, polar_type_option),
            promotes=["*"],
        )



"""
Estimation of wing wet area
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from ..constants import SERVICE_WING_GEOMETRY_WET_AREA


@oad.RegisterSubmodel(
    SERVICE_WING_GEOMETRY_WET_AREA, "fastoad.submodel.geometry.wing.wet_area.legacy"
)
class ComputeWetAreaWing(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing wet area estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")

        self.add_output("data:geometry:wing:outer_area", units="m**2")
        self.add_output("data:geometry:wing:wetted_area", units="m**2")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:wing:outer_area",
            [
                "data:geometry:wing:area",
                "data:geometry:wing:root:y",
                "data:geometry:wing:root:chord",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:wing:wetted_area",
            [
                "data:geometry:wing:area",
                "data:geometry:wing:root:chord",
                "data:geometry:fuselage:maximum_width",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        wing_area = inputs["data:geometry:wing:area"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        width_max = inputs["data:geometry:fuselage:maximum_width"]

        s_pf = wing_area - 2 * l2_wing * y2_wing
        wet_area_wing = 2 * (wing_area - width_max * l2_wing)

        outputs["data:geometry:wing:outer_area"] = s_pf
        outputs["data:geometry:wing:wetted_area"] = wet_area_wing



"""
Estimation of center of gravity ratio with aft
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_EMPTY_AIRCRAFT_CG


@RegisterSubmodel(SERVICE_EMPTY_AIRCRAFT_CG, "fastoad.submodel.weight.cg.empty_aircraft.legacy")
class ComputeCGRatioAft(om.Group):
    def setup(self):
        self.add_subsystem("cg_all", ComputeCG(), promotes=["*"])
        self.add_subsystem("cg_ratio", CGRatio(), promotes=["*"])


class ComputeCG(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            "cg_names",
            default=[
                "data:weight:airframe:wing:CG:x",
                "data:weight:airframe:fuselage:CG:x",
                "data:weight:airframe:horizontal_tail:CG:x",
                "data:weight:airframe:vertical_tail:CG:x",
                "data:weight:airframe:flight_controls:CG:x",
                "data:weight:airframe:landing_gear:main:CG:x",
                "data:weight:airframe:landing_gear:front:CG:x",
                "data:weight:airframe:pylon:CG:x",
                "data:weight:airframe:paint:CG:x",
                "data:weight:propulsion:engine:CG:x",
                "data:weight:propulsion:fuel_lines:CG:x",
                "data:weight:propulsion:unconsumables:CG:x",
                "data:weight:systems:power:auxiliary_power_unit:CG:x",
                "data:weight:systems:power:electric_systems:CG:x",
                "data:weight:systems:power:hydraulic_systems:CG:x",
                "data:weight:systems:life_support:insulation:CG:x",
                "data:weight:systems:life_support:air_conditioning:CG:x",
                "data:weight:systems:life_support:de-icing:CG:x",
                "data:weight:systems:life_support:cabin_lighting:CG:x",
                "data:weight:systems:life_support:seats_crew_accommodation:CG:x",
                "data:weight:systems:life_support:oxygen:CG:x",
                "data:weight:systems:life_support:safety_equipment:CG:x",
                "data:weight:systems:navigation:CG:x",
                "data:weight:systems:transmission:CG:x",
                "data:weight:systems:operational:radar:CG:x",
                "data:weight:systems:operational:cargo_hold:CG:x",
                "data:weight:systems:flight_kit:CG:x",
                "data:weight:furniture:passenger_seats:CG:x",
                "data:weight:furniture:food_water:CG:x",
                "data:weight:furniture:security_kit:CG:x",
                "data:weight:furniture:toilets:CG:x",
            ],
        )

        self.options.declare(
            "mass_names",
            [
                "data:weight:airframe:wing:mass",
                "data:weight:airframe:fuselage:mass",
                "data:weight:airframe:horizontal_tail:mass",
                "data:weight:airframe:vertical_tail:mass",
                "data:weight:airframe:flight_controls:mass",
                "data:weight:airframe:landing_gear:main:mass",
                "data:weight:airframe:landing_gear:front:mass",
                "data:weight:airframe:pylon:mass",
                "data:weight:airframe:paint:mass",
                "data:weight:propulsion:engine:mass",
                "data:weight:propulsion:fuel_lines:mass",
                "data:weight:propulsion:unconsumables:mass",
                "data:weight:systems:power:auxiliary_power_unit:mass",
                "data:weight:systems:power:electric_systems:mass",
                "data:weight:systems:power:hydraulic_systems:mass",
                "data:weight:systems:life_support:insulation:mass",
                "data:weight:systems:life_support:air_conditioning:mass",
                "data:weight:systems:life_support:de-icing:mass",
                "data:weight:systems:life_support:cabin_lighting:mass",
                "data:weight:systems:life_support:seats_crew_accommodation:mass",
                "data:weight:systems:life_support:oxygen:mass",
                "data:weight:systems:life_support:safety_equipment:mass",
                "data:weight:systems:navigation:mass",
                "data:weight:systems:transmission:mass",
                "data:weight:systems:operational:radar:mass",
                "data:weight:systems:operational:cargo_hold:mass",
                "data:weight:systems:flight_kit:mass",
                "data:weight:furniture:passenger_seats:mass",
                "data:weight:furniture:food_water:mass",
                "data:weight:furniture:security_kit:mass",
                "data:weight:furniture:toilets:mass",
            ],
        )

    def setup(self):
        for cg_name in self.options["cg_names"]:
            self.add_input(cg_name, val=np.nan, units="m")
        for mass_name in self.options["mass_names"]:
            self.add_input(mass_name, val=np.nan, units="kg")

        self.add_output("data:weight:aircraft_empty:mass", units="kg")
        self.add_output("data:weight:aircraft_empty:CG:x", units="m")

    def setup_partials(self):
        self.declare_partials("data:weight:aircraft_empty:mass", "*", method="fd")
        self.declare_partials("data:weight:aircraft_empty:CG:x", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        cgs = [inputs[cg_name][0] for cg_name in self.options["cg_names"]]
        masses = [inputs[mass_name][0] for mass_name in self.options["mass_names"]]

        weight_moment = np.dot(cgs, masses)
        outputs["data:weight:aircraft_empty:mass"] = np.sum(masses)
        outputs["data:weight:aircraft_empty:CG:x"] = (
            weight_moment / outputs["data:weight:aircraft_empty:mass"]
        )


class CGRatio(om.ExplicitComponent):
    def setup(self):
        self.add_input("data:weight:aircraft_empty:CG:x", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")

        self.add_output("data:weight:aircraft:empty:CG:MAC_position")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        x_cg_all = inputs["data:weight:aircraft_empty:CG:x"]
        wing_position = inputs["data:geometry:wing:MAC:at25percent:x"]
        mac = inputs["data:geometry:wing:MAC:length"]

        outputs["data:weight:aircraft:empty:CG:MAC_position"] = (
            x_cg_all - wing_position + 0.25 * mac
        ) / mac



"""
Airfoil reshape function
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from importlib.resources import open_text

import numpy as np

from . import resources
from .profile import Profile


# FIXME: user currently has to know the name of available resources. A class would allow
#        to provide more introspection.
def get_profile(file_name: str = "BACJ.txt", chord_length=1.0, thickness_ratio=None) -> Profile:
    """
    Reads profile from indicated resource file and returns it after resize

    :param file_name: name of resource
    :param chord_length: set to None to get original chord length
    :param thickness_ratio:
    :return: the Profile instance
    """

    with open_text(resources, file_name) as source:
        x_z = np.genfromtxt(source, skip_header=1, delimiter="\t", names="x, z")
    profile = Profile()
    profile.set_points(x_z["x"], x_z["z"])

    if thickness_ratio:
        profile.thickness_ratio = thickness_ratio

    if chord_length:
        profile.chord_length = chord_length

    return profile



"""
Estimation of geometry of fuselase part A - Cabin (Commercial)
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from math import sqrt

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from ...constants import (
    SERVICE_FUSELAGE_GEOMETRY_BASIC,
    SERVICE_FUSELAGE_GEOMETRY_WITH_CABIN_SIZING,
)


@oad.RegisterSubmodel(
    SERVICE_FUSELAGE_GEOMETRY_BASIC, "fastoad.submodel.geometry.fuselage.basic.legacy"
)
class ComputeFuselageGeometryBasic(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Geometry of fuselage part A - Cabin (Commercial) estimation"""

    def setup(self):
        self.add_input("data:geometry:cabin:NPAX1", val=np.nan)
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:front_length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:rear_length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:PAX_length", val=np.nan, units="m")

        self.add_output("data:weight:systems:flight_kit:CG:x", units="m")
        self.add_output("data:weight:furniture:passenger_seats:CG:x", units="m")
        self.add_output("data:geometry:cabin:length", units="m")
        self.add_output("data:geometry:fuselage:wetted_area", units="m**2")
        self.add_output("data:geometry:cabin:crew_count:commercial")

    def setup_partials(self):
        self.declare_partials(
            "data:weight:systems:flight_kit:CG:x",
            ["data:geometry:fuselage:front_length", "data:geometry:fuselage:PAX_length"],
            method="fd",
        )
        self.declare_partials(
            "data:weight:furniture:passenger_seats:CG:x",
            ["data:geometry:fuselage:front_length", "data:geometry:fuselage:PAX_length"],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:cabin:length", ["data:geometry:fuselage:length"], method="fd"
        )
        self.declare_partials(
            "data:geometry:fuselage:wetted_area",
            [
                "data:geometry:fuselage:maximum_width",
                "data:geometry:fuselage:maximum_height",
                "data:geometry:fuselage:front_length",
                "data:geometry:fuselage:rear_length",
                "data:geometry:fuselage:length",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:cabin:crew_count:commercial", ["data:geometry:cabin:NPAX1"], method="fd"
        )

    def compute(self, inputs, outputs):
        npax_1 = inputs["data:geometry:cabin:NPAX1"]
        fus_length = inputs["data:geometry:fuselage:length"]
        b_f = inputs["data:geometry:fuselage:maximum_width"]
        h_f = inputs["data:geometry:fuselage:maximum_height"]
        lav = inputs["data:geometry:fuselage:front_length"]
        lar = inputs["data:geometry:fuselage:rear_length"]
        lpax = inputs["data:geometry:fuselage:PAX_length"]

        l_cyl = fus_length - lav - lar
        cabin_length = 0.81 * fus_length
        x_cg_d2 = lav + 0.35 * lpax
        x_cg_c6 = lav + 0.1 * lpax
        pnc = int((npax_1 + 17) / 35)

        # Equivalent diameter of the fuselage
        fus_dia = sqrt(b_f * h_f)
        wet_area_nose = 2.45 * fus_dia * lav
        wet_area_cyl = 3.1416 * fus_dia * l_cyl
        wet_area_tail = 2.3 * fus_dia * lar
        wet_area_fus = wet_area_nose + wet_area_cyl + wet_area_tail

        outputs["data:weight:systems:flight_kit:CG:x"] = x_cg_c6
        outputs["data:weight:furniture:passenger_seats:CG:x"] = x_cg_d2
        outputs["data:geometry:cabin:length"] = cabin_length
        outputs["data:geometry:cabin:crew_count:commercial"] = pnc
        outputs["data:geometry:fuselage:wetted_area"] = wet_area_fus


@oad.RegisterSubmodel(
    SERVICE_FUSELAGE_GEOMETRY_WITH_CABIN_SIZING, "geometry.fuselage.with_cabin_sizing.legacy"
)
class ComputeFuselageGeometryCabinSizing(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Geometry of fuselage part A - Cabin (Commercial) estimation"""

    def setup(self):
        self.add_input("data:geometry:cabin:seats:economical:width", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:seats:economical:length", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:seats:economical:count_by_row", val=np.nan)
        self.add_input("data:geometry:cabin:aisle_width", val=np.nan, units="m")
        self.add_input("data:geometry:cabin:exit_width", val=np.nan, units="m")
        self.add_input("data:TLAR:NPAX", val=np.nan)
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)

        self.add_output("data:geometry:cabin:NPAX1")
        self.add_output("data:weight:systems:flight_kit:CG:x", units="m")
        self.add_output("data:weight:furniture:passenger_seats:CG:x", units="m")
        self.add_output("data:geometry:fuselage:length", units="m")
        self.add_output("data:geometry:fuselage:maximum_width", units="m")
        self.add_output("data:geometry:fuselage:maximum_height", units="m")
        self.add_output("data:geometry:fuselage:front_length", units="m")
        self.add_output("data:geometry:fuselage:rear_length", units="m")
        self.add_output("data:geometry:fuselage:PAX_length", units="m")
        self.add_output("data:geometry:cabin:length", units="m")
        self.add_output("data:geometry:fuselage:wetted_area", units="m**2")
        self.add_output("data:geometry:cabin:crew_count:commercial")

    def setup_partials(self):
        self.declare_partials("data:geometry:cabin:NPAX1", ["data:TLAR:NPAX"], method="fd")
        self.declare_partials(
            "data:geometry:fuselage:maximum_width",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:maximum_height",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:front_length",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:rear_length",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:PAX_length",
            ["data:geometry:cabin:seats:economical:length", "data:geometry:cabin:exit_width"],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:length",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:aisle_width",
                "data:geometry:cabin:seats:economical:length",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:exit_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:cabin:length",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:aisle_width",
                "data:geometry:cabin:seats:economical:length",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:exit_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:weight:systems:flight_kit:CG:x",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:seats:economical:length",
                "data:geometry:cabin:exit_width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:weight:furniture:passenger_seats:CG:x",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:seats:economical:length",
                "data:geometry:cabin:exit_width",
                "data:geometry:cabin:aisle_width",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:fuselage:wetted_area",
            [
                "data:geometry:cabin:seats:economical:count_by_row",
                "data:geometry:cabin:aisle_width",
                "data:geometry:cabin:seats:economical:length",
                "data:geometry:cabin:seats:economical:width",
                "data:geometry:cabin:exit_width",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        front_seat_number_eco = inputs["data:geometry:cabin:seats:economical:count_by_row"]
        ws_eco = inputs["data:geometry:cabin:seats:economical:width"]
        ls_eco = inputs["data:geometry:cabin:seats:economical:length"]
        w_aisle = inputs["data:geometry:cabin:aisle_width"]
        w_exit = inputs["data:geometry:cabin:exit_width"]
        npax = inputs["data:TLAR:NPAX"]
        n_engines = inputs["data:geometry:propulsion:engine:count"]

        # Cabin width = N * seat width + Aisle width + (N+2)*2"+2 * 1"
        wcabin = (
            front_seat_number_eco * ws_eco + w_aisle + (front_seat_number_eco + 2) * 0.051 + 0.05
        )

        # Number of rows = Npax / N
        npax_1 = int(1.05 * npax)
        n_rows = int(npax_1 / front_seat_number_eco)
        pnc = int((npax + 17) / 35)
        # Length of pax cabin = Length of seat area + Width of 1 Emergency
        # exits
        lpax = (n_rows * ls_eco) + 1 * w_exit
        l_cyl = lpax - (2 * front_seat_number_eco - 4) * ls_eco
        r_i = wcabin / 2
        radius = 1.06 * r_i
        # Cylindrical fuselage
        b_f = 2 * radius
        # 0.14m is the distance between both lobe centers of the fuselage
        h_f = b_f + 0.14
        lav = 1.7 * h_f

        if n_engines == 3.0:
            lar = 3.0 * h_f
        else:
            lar = 3.60 * h_f

        fus_length = lav + lar + l_cyl
        cabin_length = 0.81 * fus_length
        x_cg_c6 = lav - (front_seat_number_eco - 4) * ls_eco + lpax * 0.1
        x_cg_d2 = lav - (front_seat_number_eco - 4) * ls_eco + lpax / 2

        # Equivalent diameter of the fuselage
        fus_dia = sqrt(b_f * h_f)
        wet_area_nose = 2.45 * fus_dia * lav
        wet_area_cyl = 3.1416 * fus_dia * l_cyl
        wet_area_tail = 2.3 * fus_dia * lar
        wet_area_fus = wet_area_nose + wet_area_cyl + wet_area_tail

        outputs["data:geometry:cabin:NPAX1"] = npax_1
        outputs["data:weight:systems:flight_kit:CG:x"] = x_cg_c6
        outputs["data:weight:furniture:passenger_seats:CG:x"] = x_cg_d2
        outputs["data:geometry:fuselage:length"] = fus_length
        outputs["data:geometry:fuselage:maximum_width"] = b_f
        outputs["data:geometry:fuselage:maximum_height"] = h_f
        outputs["data:geometry:fuselage:front_length"] = lav
        outputs["data:geometry:fuselage:rear_length"] = lar
        outputs["data:geometry:fuselage:PAX_length"] = lpax
        outputs["data:geometry:cabin:length"] = cabin_length
        outputs["data:geometry:cabin:crew_count:commercial"] = pnc
        outputs["data:geometry:fuselage:wetted_area"] = wet_area_fus



"""
Estimation of main landing gear center of gravity
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_MLG_CG


@RegisterSubmodel(SERVICE_MLG_CG, "fastoad.submodel.weight.cg.main_landing_gear")
class UpdateMLG(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Main landing gear center of gravity estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:weight:aircraft:CG:aft:MAC_position", val=np.nan)
        self.add_input("data:weight:airframe:landing_gear:front:CG:x", units="m")
        self.add_input("settings:weight:airframe:landing_gear:front:weight_ratio", val=0.08)

        self.add_output("data:weight:airframe:landing_gear:main:CG:x", units="m")

    def setup_partials(self):
        self.declare_partials("data:weight:airframe:landing_gear:main:CG:x", "*", method="fd")

    def compute(self, inputs, outputs):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        cg_ratio = inputs["data:weight:aircraft:CG:aft:MAC_position"]
        cg_a52 = inputs["data:weight:airframe:landing_gear:front:CG:x"]
        front_lg_weight_ratio = inputs["settings:weight:airframe:landing_gear:front:weight_ratio"]

        x_cg = fa_length - 0.25 * l0_wing + cg_ratio * l0_wing

        cg_a51 = (x_cg - front_lg_weight_ratio * cg_a52) / (1 - front_lg_weight_ratio)

        outputs["data:weight:airframe:landing_gear:main:CG:x"] = cg_a51



"""Computation of form drag for whole aircraft."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import (
    SERVICE_CD0,
    SERVICE_CD0_FUSELAGE,
    SERVICE_CD0_HORIZONTAL_TAIL,
    SERVICE_CD0_NACELLES_PYLONS,
    SERVICE_CD0_SUM,
    SERVICE_CD0_VERTICAL_TAIL,
    SERVICE_CD0_WING,
)


@RegisterSubmodel(SERVICE_CD0, "fastoad.submodel.aerodynamics.CD0.legacy")
class CD0(om.Group):
    """
    Computation of form drag for whole aircraft.

    Computes and sums the drag coefficients of all components.
    Interaction drag is assumed to be taken into account at component level.
    """

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        low_speed_option = {"low_speed_aero": self.options["low_speed_aero"]}
        self.add_subsystem(
            "cd0_wing",
            RegisterSubmodel.get_submodel(SERVICE_CD0_WING, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_fuselage",
            RegisterSubmodel.get_submodel(SERVICE_CD0_FUSELAGE, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_ht",
            RegisterSubmodel.get_submodel(SERVICE_CD0_HORIZONTAL_TAIL, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_vt",
            RegisterSubmodel.get_submodel(SERVICE_CD0_VERTICAL_TAIL, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_nac_pylons",
            RegisterSubmodel.get_submodel(SERVICE_CD0_NACELLES_PYLONS, low_speed_option),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_total",
            RegisterSubmodel.get_submodel(SERVICE_CD0_SUM, low_speed_option),
            promotes=["*"],
        )



"""
Estimation of aerodynamic center
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from .constants import SERVICE_AIRCRAFT_AERODYNAMIC_CENTER


@oad.RegisterSubmodel(
    SERVICE_AIRCRAFT_AERODYNAMIC_CENTER,
    "fastoad.submodel.geometry.aircraft.aerodynamic_center.legacy",
)
class ComputeAeroCenter(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Aerodynamic center estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:MAC:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:virtual_chord", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:horizontal_tail:area", val=np.nan, units="m**2")
        self.add_input(
            "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25", val=np.nan, units="m"
        )
        self.add_input("data:aerodynamics:aircraft:cruise:CL_alpha", val=np.nan, units="1/rad")
        self.add_input(
            "data:aerodynamics:horizontal_tail:cruise:CL_alpha", val=np.nan, units="1/rad"
        )

        self.add_output("data:aerodynamics:cruise:neutral_point:x")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        x0_wing = inputs["data:geometry:wing:MAC:leading_edge:x:local"]
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        l1_wing = inputs["data:geometry:wing:root:virtual_chord"]
        width_max = inputs["data:geometry:fuselage:maximum_width"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        fus_length = inputs["data:geometry:fuselage:length"]
        wing_area = inputs["data:geometry:wing:area"]
        s_h = inputs["data:geometry:horizontal_tail:area"]
        lp_ht = inputs["data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25"]
        cl_alpha_wing = inputs["data:aerodynamics:aircraft:cruise:CL_alpha"]
        cl_alpha_ht = inputs["data:aerodynamics:horizontal_tail:cruise:CL_alpha"]
        # TODO: make variable name is computation sequence more english
        x0_25 = fa_length - 0.25 * l0_wing - x0_wing + 0.25 * l1_wing
        ratio_x025 = x0_25 / fus_length
        # fitting result of Raymer book, figure 16.14
        k_h = 0.01222 - 7.40541e-4 * ratio_x025 * 100 + 2.1956e-5 * (ratio_x025 * 100) ** 2
        # equation from Raymer book, eqn 16.22
        cm_alpha_fus = k_h * width_max**2 * fus_length / (l0_wing * wing_area) * 180.0 / np.pi
        x_ca_plane = (
            cl_alpha_wing * fa_length / l0_wing
            - cm_alpha_fus
            + cl_alpha_ht * (1 - 0.4) * 0.9 * s_h / wing_area * (lp_ht + fa_length) / l0_wing
        ) / (cl_alpha_wing + cl_alpha_ht * (1 - 0.4) * 0.9 * s_h / wing_area)
        x_aero_center = x_ca_plane - fa_length / l0_wing + 0.25

        outputs["data:aerodynamics:cruise:neutral_point:x"] = x_aero_center



"""OpenMDAO wrapping of RubberEngine."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.model_base.propulsion import (
    BaseOMPropulsionComponent,
    FuelEngineSet,
    IOMPropulsionWrapper,
    IPropulsion,
)
from fastoad.module_management.service_registry import RegisterPropulsion
from fastoad.openmdao.validity_checker import ValidityDomainChecker
from openmdao.core.component import Component

from .constants import RUBBER_ENGINE_DESCRIPTION
from .rubber_engine import RubberEngine


@RegisterPropulsion("fastoad.wrapper.propulsion.rubber_engine", desc=RUBBER_ENGINE_DESCRIPTION)
class OMRubberEngineWrapper(IOMPropulsionWrapper):
    """
    Wrapper class of for rubber engine model.

    It is made to allow a direct call to
    :class:`~fastoad.models.propulsion.fuel_propulsion.rubber_engine.rubber_engine.RubberEngine`
    in an OpenMDAO component.

    Example of usage of this class::

        import openmdao.api as om

        class MyComponent(om.ExplicitComponent):
            def initialize():
                self._engine_wrapper = OMRubberEngineWrapper()

            def setup():
                # Adds OpenMDAO variables that define the engine
                self._engine_wrapper.setup(self)

                # Do the normal setup
                self.add_input("my_input")
                [finish the setup...]

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                [do something]

                # Get the engine instance, with parameters defined from OpenMDAO inputs
                engine = self._engine_wrapper.get_model(inputs)

                # Run the engine model. This is a pure Python call. You have to define
                # its inputs before, and to use its outputs according to your needs
                sfc, thrust_rate, thrust = engine.compute_flight_points(
                    mach,
                    altitude,
                    engine_setting,
                    use_thrust_rate,
                    thrust_rate,
                    thrust
                    )

                [do something else]

        )
    """

    def setup(self, component: Component):
        component.add_input("data:propulsion:rubber_engine:bypass_ratio", np.nan)
        component.add_input("data:propulsion:rubber_engine:overall_pressure_ratio", np.nan)
        component.add_input(
            "data:propulsion:rubber_engine:turbine_inlet_temperature", np.nan, units="K"
        )
        component.add_input("data:propulsion:MTO_thrust", np.nan, units="N")
        component.add_input("data:propulsion:rubber_engine:maximum_mach", np.nan)
        component.add_input("data:propulsion:rubber_engine:design_altitude", np.nan, units="m")
        component.add_input(
            "data:propulsion:rubber_engine:delta_t4_climb",
            -50,
            desc="As it is a delta, unit is K or °C, but is not "
            "specified to avoid OpenMDAO making unwanted conversion",
        )
        component.add_input(
            "data:propulsion:rubber_engine:delta_t4_cruise",
            -100,
            desc="As it is a delta, unit is K or °C, but is not "
            "specified to avoid OpenMDAO making unwanted conversion",
        )
        component.add_input("tuning:propulsion:rubber_engine:SFC:k_sl", 1.0)
        component.add_input("tuning:propulsion:rubber_engine:SFC:k_cr", 1.0)
        component.add_input("data:geometry:propulsion:engine:count", 2)

    @staticmethod
    def get_model(inputs) -> IPropulsion:
        """

        :param inputs: input parameters that define the engine
        :return: a RubberEngine instance
        """
        engine_params = {
            "bypass_ratio": inputs["data:propulsion:rubber_engine:bypass_ratio"],
            "overall_pressure_ratio": inputs[
                "data:propulsion:rubber_engine:overall_pressure_ratio"
            ],
            "turbine_inlet_temperature": inputs[
                "data:propulsion:rubber_engine:turbine_inlet_temperature"
            ],
            "maximum_mach": inputs["data:propulsion:rubber_engine:maximum_mach"],
            "design_altitude": inputs["data:propulsion:rubber_engine:design_altitude"],
            "delta_t4_climb": inputs["data:propulsion:rubber_engine:delta_t4_climb"],
            "delta_t4_cruise": inputs["data:propulsion:rubber_engine:delta_t4_cruise"],
            "mto_thrust": inputs["data:propulsion:MTO_thrust"],
            "k_sfc_sl": inputs["tuning:propulsion:rubber_engine:SFC:k_sl"],
            "k_sfc_cr": inputs["tuning:propulsion:rubber_engine:SFC:k_cr"],
        }

        return FuelEngineSet(
            RubberEngine(**engine_params), inputs["data:geometry:propulsion:engine:count"]
        )


@ValidityDomainChecker(
    {
        "data:propulsion:altitude": (None, 20000.0),
        "data:propulsion:mach": (0.75, 0.85),  # limitation of SFC ratio model
        "data:propulsion:rubber_engine:overall_pressure_ratio": (20.0, 40.0),
        "data:propulsion:rubber_engine:bypass_ratio": (3.0, 6.0),
        "data:propulsion:thrust_rate": (0.5, 1.0),  # limitation of SFC ratio model
        "data:propulsion:rubber_engine:turbine_inlet_temperature": (
            1400.0,
            1600.0,
        ),  # limitation of max thrust model
        "data:propulsion:rubber_engine:delta_t4_climb": (
            -100.0,
            0.0,
        ),  # limitation of max thrust model
        "data:propulsion:rubber_engine:delta_t4_cruise": (
            -100.0,
            0.0,
        ),  # limitation of max thrust model
    }
)
# @RegisterOpenMDAOSystem(
#     "fastoad.propulsion.rubber_engine",
#     desc=RUBBER_ENGINE_DESCRIPTION,
#     domain=ModelDomain.PROPULSION,
# )
class OMRubberEngineComponent(BaseOMPropulsionComponent):
    """
    Parametric engine model as OpenMDAO component

    See
    :class:`~fastoad.models.propulsion.fuel_propulsion.rubber_engine.rubber_engine.RubberEngine`
    for more information.
    """

    def setup(self):
        super().setup()
        self.get_wrapper().setup(self)

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    @staticmethod
    def get_wrapper() -> OMRubberEngineWrapper:
        return OMRubberEngineWrapper()



"""
Estimation of fuel lines weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_FUEL_LINES_MASS


@RegisterSubmodel(
    SERVICE_FUEL_LINES_MASS, "fastoad.submodel.weight.mass.propulsion.fuel_lines.legacy"
)
class FuelLinesWeight(om.ExplicitComponent):
    """
    Weight estimation for fuel lines

    Based on formula in :cite:`supaero:2014`, mass contribution B2
    """

    def setup(self):
        self.add_input("data:geometry:wing:b_50", val=np.nan, units="m")
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")
        self.add_input("data:weight:propulsion:engine:mass", val=np.nan, units="kg")
        self.add_input("tuning:weight:propulsion:fuel_lines:mass:k", val=1.0)
        self.add_input("tuning:weight:propulsion:fuel_lines:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:propulsion:fuel_lines:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        b_50 = inputs["data:geometry:wing:b_50"]
        mfw = inputs["data:weight:aircraft:MFW"]
        k_b2 = inputs["tuning:weight:propulsion:fuel_lines:mass:k"]
        offset_b2 = inputs["tuning:weight:propulsion:fuel_lines:mass:offset"]
        weight_engines = inputs["data:weight:propulsion:engine:mass"]

        temp_b2 = 0.02 * weight_engines + 2.0 * b_50 + 0.35 * mfw**0.66
        outputs["data:weight:propulsion:fuel_lines:mass"] = k_b2 * temp_b2 + offset_b2



"""
Estimation of flight controls weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_FLIGHT_CONTROLS_MASS


@RegisterSubmodel(
    SERVICE_FLIGHT_CONTROLS_MASS, "fastoad.submodel.weight.mass.airframe.flight_control.legacy"
)
class FlightControlsWeight(om.ExplicitComponent):
    """
    Flight controls weight estimation

    Based on formulas in :cite:`supaero:2014`, mass contribution A4
    """

    def setup(self):
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:b_50", val=np.nan, units="m")
        self.add_input("data:mission:sizing:cs25:sizing_load_1", val=np.nan, units="kg")
        self.add_input("data:mission:sizing:cs25:sizing_load_2", val=np.nan, units="kg")
        self.add_input(
            "settings:weight:airframe:flight_controls:mass:k_fc", val=0.85
        )  # FIXME: this one should depend on a boolan electric/not-electric flight_controls
        self.add_input("tuning:weight:airframe:flight_controls:mass:k", val=1.0)
        self.add_input("tuning:weight:airframe:flight_controls:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:airframe:flight_controls:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        fus_length = inputs["data:geometry:fuselage:length"]
        b_50 = inputs["data:geometry:wing:b_50"]
        k_fc = inputs["settings:weight:airframe:flight_controls:mass:k_fc"]
        k_a4 = inputs["tuning:weight:airframe:flight_controls:mass:k"]
        offset_a4 = inputs["tuning:weight:airframe:flight_controls:mass:offset"]

        max_nm = max(
            inputs["data:mission:sizing:cs25:sizing_load_1"],
            inputs["data:mission:sizing:cs25:sizing_load_2"],
        )

        temp_a4 = k_fc * max_nm * (fus_length**0.66 + b_50**0.66)

        outputs["data:weight:airframe:flight_controls:mass"] = k_a4 * temp_a4 + offset_a4
